/*
 *      Copyright 2006-2006 Richard Gorton - All rights reserved
 *      Exclusively licensed to Gorton Studios Inc.
 *
 * dirwalk.c - routines to walk through a Linux directory & all subdirectories.
 *
 *	Stuff to flag:
 *		empty directories
 *		directories with bad permissions
 *              symbolic links to directories
 */

#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>

#include "wslc.h"

/*
 * Recursively traverse the directory starting at name.
 * order of the walk is not relevant, so this can be trivially recursive
 */

void RecurseWalkDir (const char *name)
{
    const char *func = "RecurseWalkDir";
    DIR	*dirstream;
    struct dirent       *dir_e;
    char        fullpath[PATH_MAX];
    char        dirpath[PATH_MAX];
    char        filepath[PATH_MAX];
    char        *fp_c;
    char        *dp_c;
    struct stat st_buf;
    char        errbuf[128];

    dp_c = realpath (name, dirpath);
    if (!dp_c) {
        InternalError (func, 1, IERR_REALPATH);
        return;
    }
    dirstream = opendir (dirpath);
    if (!dirstream) {
        snprintf (errbuf, sizeof(errbuf),
                  "%s: failure opening directory [%s] errno = %d",
                  func, dirpath, errno);
        ReportSiteError (errbuf);
        goto done;
    }

        /*
         * Valid directory, walk through it
         * regular files, and links to regular (non-directories) first
         * recursively walk subdirectories
         * Flag all symbolic links to non-regular files as errors.
         * What to do with symbolic links?  Is it possible to handle them
         * without a pile of effort?  Consider a symlink to ../.. and
         * then traversing (infinite loop)
         * Hmmm - websites probably don't use symbolic links to directories,
         * so lets avoid the extra "was I here already" overhead and simply
         * report links as errors to be investigated.
         * 9-Feb-2007 Drew: sym links do get used, but not to directories within
         * a tree.  They are used to get to DocRoot, though
         */
    dir_e = readdir (dirstream);
    while (dir_e) {
        snprintf (filepath, sizeof(filepath), "%s/%s", dirpath, dir_e->d_name);
        fp_c = realpath ((const char *)filepath, fullpath);
        if (!fp_c) {
            snprintf (errbuf, sizeof(errbuf), "%s: realpath failure, errno: %d",
                    func, errno);
            ReportSiteError (errbuf);
            goto nextreg;
        }
        switch (dir_e->d_type) {
            case DT_DIR:
                break;
            case DT_LNK: {
                stat (fullpath, &st_buf);
                if (!S_ISREG(st_buf.st_mode)) {
                    snprintf (errbuf, sizeof(errbuf), "%s: link [%s] is not to a regular file",
                            func, fullpath);
                    ReportSiteError (errbuf);
                    goto nextreg;
                }
                // else fall into regular file code
                break;
            }
            case DT_REG: {
                AddFileToFullFileList (fullpath);
                break;
            }
            default: {
                // Weird file - flag as an error
                snprintf (errbuf, sizeof(errbuf),
                        "%s: abnormal file type [%s]: %d (see <dirent.h>)",
                        func, fullpath, dir_e->d_type);
                ReportSiteError (errbuf);
                break;
            }
        }       // switch
nextreg:
        dir_e = readdir (dirstream);
    }

        // all subdirectories
    rewinddir (dirstream);
    dir_e = readdir (dirstream);
    while (dir_e) {
        if (dir_e->d_type == DT_DIR) {
            snprintf (filepath, sizeof(filepath), "%s/%s", dirpath, dir_e->d_name);
            fp_c = realpath ((const char *)filepath, fullpath);
            if (!fp_c)
                continue;
                // Skip "." and ".."
            if (dir_e->d_name[0] == '.') {
                if (dir_e->d_name[1] == 0x0) {
                    goto nextfile;
                } else if ((dir_e->d_name[1] == '.') && (dir_e->d_name[2] == 0x0)) {
                    goto nextfile;
                }
            } else {
                RecurseWalkDir (fp_c);
            }
        }       // a subdirectory
nextfile: ;
        dir_e = readdir (dirstream);
    }   // more files in this directory
    closedir (dirstream);

done: ;
}

