/*
 *      Copyright 2006-2006 Richard Gorton - All rights reserved
 *      Exclusively licensed to Gorton Studios Inc.
 *
 * wslcan.c - Web Site Link Checker and Analyzer
 *
 * Given an starting .html file, it will walk throught he on-disk web site and
 * report various hard-to-find errors.
 */
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <ctype.h>

#include "wslc.h"

HFLIST *HF_l = NULL;
FILREC *HTable[NCHAINS];
int     HTableCount = 0;
FILEDATA *ExRefs = NULL;
FILEDATA *NonAscii = NULL;
FILEDATA *NonExists = NULL;
FILEDATA *Bad_LT_A = NULL;
FILEDATA *UnknownType = NULL;

static void usage (char *me) {
    fprintf (stderr, "%s <start_html>\n", me);
    fprintf (stderr, "<start_html> is the initial html file for the web site\n");
}

// Parser assist macros and routines
#define REQUIRE_CHAR_TOK(c_p, t) \
    if (*c_p != t) { \
	char eb[64]; \
	snprintf (eb, sizeof(eb), "Error: missing %s at string %s...", t, c_p); \
	Error (eb); \
	}

static char *SkipWhiteSpace (char *c, int *lineno)
{
    while (isspace (*c)) {
        if (*c == '\n')
            *lineno += 1;
        c++;
    }
    return c;
}
/*
 * Assumes starting point is whitespace after "<a name"
 * or "<a href"
 * Returns the new scan start point after a 'name = "foo"' (quotes optional)
 */
static char *GetNameOrHref (char *start_s, char *buf, int *lineno)
{
    int	has_quote;
    int	anch_i;

    start_s = SkipWhiteSpace (start_s, lineno);
    REQUIRE_CHAR_TOK (start_s, '=');
    start_s++;
    start_s = SkipWhiteSpace (start_s, lineno);
    has_quote = (*start_s == '"');
    anch_i = 0;
    if (has_quote) {
	start_s++;	// Skip leading quote
        while ((anch_i < PATH_MAX) && (*start_s != '"')) {
	    buf[anch_i] = *start_s;
	    anch_i++;
	    start_s++;
        }
	start_s++;	// Skip trailing quote
    } else {
        // Go up to a space (might be name=foo href=...> vs. name=foo>
        while ((anch_i < PATH_MAX) && !isspace(*start_s) && (*start_s != '>')) {
	    buf[anch_i] = *start_s;
	    anch_i++;
	    start_s++;
        }
    }
        // Strip trailing whitespace
    while (isspace(buf[anch_i])) {
        anch_i--;
    }
    buf[anch_i] = 0x0;
        // buf is now a null terminated string

    return start_s;
}

/*
 * Anchor use or def happened in file hfile, with a path of path#Aname
 */
static void RecordAnchorName (char *hfile, char *path, char *AName, int defref, int *lineno)
{
    const char  *func = "RecordAnchorName";
    FILREC	*FR;
    ANL		*AnL;

    FR = FindRecByName (path);
    if (!FR) {
	char	ebuf[PATH_MAX];

	snprintf (ebuf, sizeof(ebuf) - 1, "%s referenced non-existant anchor/link at line %d: %s#%s",
			hfile, *lineno, path, AName);
	Error (ebuf);
	goto done;
    } else {
	ANL	*walker;

	walker = FR->AncList;
	while (walker) {
	    if (strcasecmp (walker->aname, AName) == 0) {
		walker->defref |= defref;
		goto done;
	    }
	    walker = walker->next;
	}

	// No matches, create listitem
	AnL = (ANL *) malloc (sizeof (*AnL));
        if (!AnL) {
            InternalError (func, 1, IERR_MALLOC);
            goto done;
        }
	strcpy (AnL->aname, AName);
	AnL->defref = defref;
	AnL->next = FR->AncList;
	FR->AncList = AnL;
    }

done: ;
}
static void RecordExternalHref (char *hfile, int *lineno, char *HRefPath, int defref)
{
    AddFileDatum (&ExRefs, hfile, lineno, HRefPath);
}

/*
 * Assumes cur_s is at the space after "<img"
 */
char *Process_lt_img (char *cur_s, char *starthtm, int *lineno)
{
    char        *last_slash;

    cur_s = SkipWhiteSpace (cur_s, lineno);
    if (strncasecmp (cur_s, "src", 3) == 0) {
	/*
	 * <img src = foo >
	 * <img src = foo alt="alt text">
	 * <img src = "foo" >
	 * <img src = "foo" alt = "alt text">
	 */
	char	ImgPath[PATH_MAX];
	char	FullPath[PATH_MAX];

	cur_s += 3;
        cur_s = GetNameOrHref (cur_s, ImgPath, lineno);
	strcpy (FullPath, starthtm);
        last_slash = rindex (FullPath, '/');
        last_slash++;
	strcpy (last_slash, ImgPath);
        AddFileToReferencedList (starthtm, lineno, FullPath);
    }

not_lt_img_src:
    return cur_s;
}

/*
 * Assumes cur_s is at the space after "<a"
 *
 * Usage cases to consider:
 *
 * Name anchor:
 *	<a name = anchor>
 *	<a name = "foo" >
 * One book claims that quotes are only required if the
 * anchor name is more than one word.  In reality, browsers
 * only like the first word (up to white space)
 *
 * HREF link:
 *	<a href = foo >
 *	<a href = "foo" >
 *	<a href = #anchor >
 *	<a href = foo.html#anchor >
 * HREF and anchor
 *	<a name = anchor href=foo.html>
 *	<a href=foo.html name = anchor>
 *	<a href=foo.html#otheranchor name = thisanchor>
 *
 * Summary: once the <a is seen, assume multiple names and hrefs are
 *	possible until the closing '>' is seen.
 *
 */
char *Process_lt_a (char *cur_s, char *starthtm, int *lineno)
{
    const char  *func = "Process_lt_a";
    char	FullPath[PATH_MAX];

    while (*cur_s != '>') {
        cur_s = SkipWhiteSpace (cur_s, lineno);
        if (strncasecmp (cur_s, "name", 4) == 0) {
            char        AnchorName[PATH_MAX];

            cur_s += 4; // end of name
            cur_s = GetNameOrHref (cur_s, AnchorName, lineno);
        	// join it with the path of the current file
            RecordAnchorName (starthtm, starthtm, AnchorName, ANCHOR_DEF, lineno);
        } else if (strncasecmp (cur_s, "href", 4) == 0) {
            char	HRefPath[PATH_MAX];
            char        *pth;

            cur_s += 4;	// end of href
            cur_s = GetNameOrHref (cur_s, HRefPath, lineno);
        	/*
        	 * Use cases:
        	 * 1) For hrefs to full http[s]://[site]/<stuff> paths, keep track
	         * of these solely as a set of refs.  There really is no way to
		 * know if [site] is the current tree.
                 * 1a) this includes href="ftp:...", href="mailto:...", and others
         	 * 2) href=[filename]
         	 * 3) href=[filename][#anchor]
         	 * 4) href=[#anchor]
         	 * Add [filename] to the list, as a REF
         	 * separate out the "#anchorname"
		 * Add anchor as a REF
		 * 
         	 */
            if ( (strncasecmp (HRefPath, "ftp://", 6) == 0) 
                || (strncasecmp (HRefPath, "http://", 7) == 0) 
                || (strncasecmp (HRefPath, "https://", 8) == 0) 
                || (strncasecmp (HRefPath, "news://", 7) == 0) 
                || (strncasecmp (HRefPath, "file://", 7) == 0) 
                || (strncasecmp (HRefPath, "mailto:", 7) == 0))
                {
                // case 1 or 1a
                RecordExternalHref (starthtm, lineno, HRefPath, ANCHOR_REF);
            } else {
                char	*pound_c;
    		char	RP[PATH_MAX];

                pound_c = rindex ((char *)HRefPath, '#');
                if (pound_c != NULL) {
			// href = [optionalstuff]#anchor
                    *pound_c = 0x0;	// Null terminate to be just a html file
                    pound_c++;
                }
		strcpy (FullPath, starthtm);
                if (pound_c != &HRefPath[1]) {
			// explicit filename given
                    char *last_slash;

                    last_slash = rindex(FullPath, '/');
                    last_slash++;
		    strcpy (last_slash, HRefPath);
		}
    		pth = realpath (FullPath, RP);
                if (!pth) {
                    int ecode = errno;

                    if (ecode == ENOENT) {
                        // The file does not exist: ref, but no def....
                        AddFileDatum (&NonExists, starthtm, lineno, FullPath);
                    } else {
                        ErrorS ("%s: realpath failure", (char *)func);
                    }
	            goto not_lt_a_space;
                }
        	AddFileToReferencedList (starthtm, lineno, RP);
                if (pound_c) {
                    RecordAnchorName (starthtm, RP, pound_c, ANCHOR_REF, lineno);
                }
            }
        } else {
        	// mailto, ftp, news, file, etc.
            char	ebuf1[16];
            char	ebuf2[64];

            snprintf (ebuf1, sizeof (ebuf1) - 1, "%s", cur_s);
            ebuf1[15] = 0x0;
            snprintf (ebuf2, sizeof (ebuf2) - 1, "[<a %s...]", cur_s, ebuf1);
            AddFileDatum (&Bad_LT_A, starthtm, lineno, ebuf2);
	    goto not_lt_a_space;
        }	// name, href, mailto
    }		// while *cur_s != '>'

not_lt_a_space:
    return cur_s;
}

static int AnalyzeHTMLFile (char *starthtm)
{
    const char  *func = "AnalyzeHTMLFile";
    int	status = 0;
    int	st_len;
    FILE	*ThisH = NULL;
    FILREC	*filrec;
    char	*FData;
    char	*cur_s;
    struct stat st_b;
    int	i;
    int		lineno;
    int         prev_nl;

    st_len = strlen (starthtm);
    filrec = FindRecByName (starthtm);
    if ((filrec->flags & FLAGS_FILE_PROCESSED)
	|| (filrec->flags & FLAGS_BEING_ANALYZED)) {
	return;
    }

    if (GetFileType (filrec) != WSLC_HTML) {
	ErrorS ("Analysis file not html: %s - aborting", starthtm);
	status = 1;
	goto done;
    }

    if (stat (starthtm, &st_b)) {
	ErrorS ("Could not stat() file: %s - aborting", starthtm);
	status = 1;
	goto done;
    }
	/*
	 * Mark this html file as being in progress to ensure
	 * cyclical recursion will not happen
	 */
    filrec->flags |= FLAGS_BEING_ANALYZED;

    ThisH = fopen (starthtm, "r");
    if (!ThisH) {
	ErrorS ("Error opening analysis root: %s - aborting", starthtm);
	status = 1;
	goto done;
    }

	/*
	 * Allocate a bit more than the actual size so we can
	 * overrun the end by a character or two with the "next '<'" math
	 */
    FData = (char *) calloc (st_b.st_size + 0x10, 1);
    if (!FData) {
        char    errbuf[256];

        sprintf (errbuf, "calloc() for [%s] in %s", starthtm, func);
        InternalError (errbuf, 1, IERR_MALLOC);
	status = 1;
	goto done;
    }

    if (fread (FData, 1, st_b.st_size, ThisH) != st_b.st_size) {
	ErrorS ("fread failure for file: %s - aborting", starthtm);
	status = 1;
	goto done;
    }

	/*
	 * Scan for and flag non-ascii content
	 */
    prev_nl = 0;
    lineno = 1;
    for (i = 0; i < st_b.st_size; i++) {
        if (FData[i] == '\n') {
            prev_nl = i;
            lineno++;
        }
	if ((FData[i] == 0x0) || (FData[i] & 0x80)) {
	    char	ebuf[1024];

	    snprintf (ebuf, sizeof(ebuf), "offset: %d value: 0x%x",
                i - prev_nl, (unsigned char) FData[i]);
            AddFileDatum (&NonAscii, starthtm, &lineno, ebuf);
	}
    }

	/*
	 * Strip out comments, converting them to spaces
	 */
    lineno = 1;
    for (i = 0; i < st_b.st_size; i++) {
	/*
	 * Comments are effectively: "<!" through ">" according to w3c.org spec
	 * Specifically:
	 * Markup declarations:
	 * <! followed by a name ending with >
	 *	<! DOCTYPE -- foobie -->
	 * Comments:
	 * <!-- stuff -->	stuff is the comment
	 * <!>			the empty comment
	 * For our purposes, convert comments to spaces
	 *
	 * BUG: context of seeing what seems to be a comment inside a string
	 * really is not.  Probably a rare case.
	 */
	if ((FData[i] == '<') && (i < st_b.st_size - 3) && (FData[i + 1] == '!')) {
	    int	c_i;

	    c_i = i;
	    while ((FData[c_i] != '>') && (c_i < st_b.st_size)) {
		FData[c_i] = 0x20;	// space
		c_i++;
	    }
	    FData[c_i] = 0x20;		// closing ">"

	// And move i up to the end of the comment
	    i = c_i + 1;
	}	// comment characters
    }	// First pass through the file


	/*
	 * constructs to evaluate:
	 *	anchors, hrefs to anchors within this file
	 *	hrefs to anchors outside this file
	 * A tag href constructs take the form:
	 *	<aZhref=stuffz>
	 *	<aZhref="stuff"z>
	 * Anchors take the form:
	 *	<a name=stuffz>
	 *	<a name="stuff"z>
	 * where Z can be any whitespace, z is empty or whitespace
	 * For our purposes, we'll assume that the aggregate format of the
	 * link has a closing </a>)
	 * Also:
	 * <a href=page10_large.jpg><IMG SRC=page10_screen.jpg></a>
	 * or even
	 * <IMG SRC=page10_screen.jpg>
	 */

    lineno = 1;
    cur_s = GetNextScanPoint (FData, &lineno);
    while (*cur_s) {
        cur_s++;
        if (tolower(*cur_s) == (char) 'a') {
	    cur_s++;
	    if (!isspace (*cur_s)) {
	        goto next_lt;
	    }
	    cur_s = Process_lt_a (cur_s, starthtm, &lineno);
        } else if (strncasecmp (cur_s, "img", 3) == 0x0) {
            cur_s += 3;
            if (!isspace (*cur_s)) {
	        goto next_lt;
	    }
            cur_s = Process_lt_img (cur_s, starthtm, &lineno);
        }
next_lt:
        cur_s = GetNextScanPoint (cur_s, &lineno);
    }

    filrec->flags |= FLAGS_FILE_PROCESSED;
    filrec->flags &= (~FLAGS_BEING_ANALYZED);

done:
    if (FData) {
	free (FData);
    }
    if (ThisH) {
        fclose (ThisH);
    }
    return status;
}

/*
 * Add [file] to the list of files found on disk
 */
int AddFileToFullFileList (char *file)
{
    const char  *func = "AddFileToFullFileList";
    FILE	*F_l;
    int		status;
    FILREC	*FR;
    long	hval;
    long	tabl_idx;
    long	i;
    char        *pth;
    char	FullPath[PATH_MAX];
    int         zero;

    status = 0;

    pth = realpath (file, FullPath);
    if (!pth) {
        status = 1;
        goto done;
    }

    FR = (FILREC *)malloc (sizeof (*FR));
    if (!FR) {
        InternalError (func, 1, IERR_MALLOC);
        status = 1;
        goto done;
    }
    strcpy (FR->filepath, FullPath);
    hval = sdbm_hash (FR->filepath, strlen(FullPath));
    FR->hash = hval;
    FR->AncList = NULL;
    FR->flags = FLAGS_UNSEEN;
    tabl_idx = hval & (NCHAINS - 1);
    FR->type = GuessFileType (FullPath, strlen(FullPath));
    if (FR->type == WSLC_UNKNOWN) {
        AddFileDatum (&UnknownType, FullPath, &zero, FullPath);
    }
    FR->next = HTable[tabl_idx];
    HTable[tabl_idx] = FR;
    HTableCount++;

done:
    return status;
}

static void Init()
{
    int i;

    for (i = 0; i < NCHAINS; i++) {
	HTable[i] = NULL;
    }
    HF_l = NULL;
}

typedef struct FILEARRAY {
    char        path[PATH_MAX];
    } FILEARRAY;
static void Fini()
{
    const char  *func = "Fini";
    int	i;
    int dindex;
    int flag;
    FILEDATA    *walker;
    FILEARRAY   *Dangling;

    ReportTypeStats();

        /*
         * Files of unknown type found during directory scanning
         */
    walker = UnknownType;
    if (walker) {
        Error ("\n\n"
               "------------------------------------------------------------"
               "\nFiles of unknown types found while walking the directories");
    }
    while (walker) {
        char    xrbuf[PATH_MAX];

        sprintf (xrbuf, "%s", walker->ht_file);
        Error (xrbuf);
        walker = walker->next;
    }

        /*
         * Report all uses of non-ascii characters
         */
    walker = NonAscii;
    if (walker) {
        Error ("\n\n"
               "------------------------------------------------------------"
               "\nNon-Ascii characters found [originating file:line offset value:]");
    }
    while (walker) {
        char    xrbuf[3*PATH_MAX];

        sprintf (xrbuf, "%s:%d %s", walker->ht_file, walker->ht_lineno, walker->ref_file);
        Error (xrbuf);
        walker = walker->next;
    }

        /*
         * Report all references to non-existant files
         */
    walker = NonExists;
    if (walker) {
        Error ("\n\n"
               "------------------------------------------------------------"
               "\nReferences to non-existing files [originating file:line file]");
    }
    while (walker) {
        char    xrbuf[3*PATH_MAX];

        sprintf (xrbuf, "%s:%d %s", walker->ht_file, walker->ht_lineno, walker->ref_file);
        Error (xrbuf);
        walker = walker->next;
    }

        /*
         * Report all external references
         */
    walker = ExRefs;
    if (walker) {
        Error ("\n\n"
               "------------------------------------------------------------"
               "\nExternal href references [originating file:line target]");
    }
    while (walker) {
        char    xrbuf[3*PATH_MAX];

        sprintf (xrbuf, "%s:%d %s", walker->ht_file, walker->ht_lineno, walker->ref_file);
        Error (xrbuf);
        walker = walker->next;
    }

        /*
         * unrecognized <a Foo... schema
         */
    walker = Bad_LT_A;
    if (walker) {
        Error ("\n\n"
               "------------------------------------------------------------"
               "\nGarbage <a schema: originating file:line [problem content]");
    }
    while (walker) {
        char    xrbuf[3*PATH_MAX];

        sprintf (xrbuf, "%s:%d %s", walker->ht_file, walker->ht_lineno, walker->ref_file);
        Error (xrbuf);
        walker = walker->next;
    }

	/*
	 * Report on files which exist but haven't been referenced.
         * Sort them first.
	 */
    Dangling = (FILEARRAY *) calloc (HTableCount, sizeof (FILEARRAY));
    if (!Dangling) {
        InternalError (func, 1, IERR_MALLOC);
        goto done;
    }
    dindex = 0;
    for (i = 0; i < NCHAINS; i++) {
	while (HTable[i]) {
	    FILREC	*next;

	    next = HTable[i]->next;
	    if (!(HTable[i]->flags & FLAGS_FILE_REFERENCED)) {
                strcpy (Dangling[dindex].path, HTable[i]->filepath);
                dindex++;
	    }
	    HTable[i] = next;
	}
    }

    if (dindex) {
        Error ("\n\n"
                "------------------------------------------------------------"
                "\nUnreferenced (dangling) files:");
        qsort ((void *)Dangling, dindex, PATH_MAX, strcmp);
    }
    for (i = 0; i < dindex; i++) {
        Error (Dangling[i].path);
    }

done:;
}

/*
 * Add this_f to the list of files referenced.
 * if it is an html file, add it to the list of html files to process
 * if we haven't already completed processing it.
 * Otherwise, simply mark it as referenced.
 */
void AddFileToReferencedList (char *starthtm, int *lineno, char *this_f)
{
    const char  *func = "AddFileToReferencedList";
    FILREC	*FR;
    HFLIST	*HEnt;
    char	RP[PATH_MAX];
    char        *pth;

    pth = realpath (this_f, RP);
    if (!pth) {
        int     ecode = errno;

        if (ecode == ENOENT) {
                // The file does not exist: ref, but no def....
            AddFileDatum (&NonExists, starthtm, lineno, this_f);
        } else {
	    ErrorS("%s: realpath failure", (char *)func);
        }
	goto done;
    }
    FR = FindRecByName (RP);
    if (!FR) {
	goto done;
    }
    FR->flags |= FLAGS_FILE_REFERENCED;
    if ((FR->type == WSLC_HTML) && !(FR->flags & FLAGS_FILE_PROCESSED)) {
	HEnt = (HFLIST *)malloc (sizeof (*HEnt));
        if (!HEnt) {
            InternalError (func, 1, IERR_MALLOC);
            goto done;
        }
	HEnt->next = (void *) HF_l;
	HEnt->FRec = FR;
	HF_l = HEnt;
    }
done: ;
}

/*
 * Get another unprocessed HTML file from the work list.
 * if the list is empty, return null.
 * Don't bother to prune the list, we might be able to use that
 * in a meaningful way later.
 */
char *GetAnotherHTML()
{
    HFLIST	*walker;
    char	*nextf;

    nextf = NULL;
    walker = HF_l;
    while (walker) {
	if (!(walker->FRec->flags & FLAGS_FILE_PROCESSED)) {
	    nextf = walker->FRec->filepath;
	    goto done;
	}
	walker = walker->next;
    }

done:
    return nextf;
}

int main (int argc, char **argv)
{
    struct stat	st_b;
    int		i;
    char	*curfil;
    char        *pth;
    char	FullPath[PATH_MAX];
    char	DirPath[PATH_MAX];

    if (argc != 2) {
	usage (argv[0]);
	exit (1);
    }

	// Validate starting point html file
    if (stat (argv[1], &st_b) || !S_ISREG(st_b.st_mode)) {
	usage (argv[0]);
	exit (1);
    }

    Init();

        // Get directory name that the start file is in
    snprintf (FullPath, sizeof(FullPath), "%s/../", argv[1]);
    FullPath [PATH_MAX] = 0x0;
    pth = realpath (FullPath, DirPath);
    if (!pth) {
        usage (argv[0]);
        exit (1);
    }

    RecurseWalkDir (DirPath);

    i = 0;
    AddFileToReferencedList (argv[1], &i, argv[1]);
    while ((curfil = GetAnotherHTML()) != NULL) {
	AnalyzeHTMLFile (curfil);
    }

    Fini();
done:
    return 0;
}
